﻿using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading;

namespace HuxBot
{
    class User
    {
        public User(string username)
        {
            this.Username = username;
            this.RequestQueue = new Queue<string>();
        }
        public User(string username, string message)
            : this(username)
        {
            this.RequestQueue.Enqueue(message);
        }
        public string Username { get; set; }
        public Queue<string> RequestQueue { get; private set; }
    }
    class SendQueue
    {
        Timer tim;
        IRCBot host;
        public bool shouldRun = false;
        public Dictionary<string, User> Users;  //Dictionary of users currently being processed
        public ConcurrentQueue<User> UserQueue; //List of order for which users should be processed
        public SendQueue(IRCBot launcher)
        {
            this.Users = new Dictionary<string, User>();
            this.UserQueue = new ConcurrentQueue<User>();
            this.tim = new Timer(WorkerTick, null, Timeout.Infinite, Timeout.Infinite);
            this.host = launcher;
        }
        public void Start()
        {
            this.tim.Change(0, 1000);
            this.shouldRun = true;
        }
        public void Add(string username, string request)
        {
            lock (this.UserQueue) //For threadsafety
            {
                if (this.Users.ContainsKey(username))
                {
                    //The user is in the user list. That means he has previously sent request that are awaiting to be processed.
                    //As such, we can safely add his new message at the end of HIS request list.

                    this.Users[username].RequestQueue.Enqueue(request); //Add users new message at the end of the list
                    return;
                }
                //User is not in the user list. Means it's his first request. Create him in the user list and add his message
                var user = new User(username, request);
                this.Users.Add(username, user); //Create the user and his message
                this.UserQueue.Enqueue(user); //Add the user to the last of the precessing users.
            }
        }
        public void WorkerTick(object sender)
        {
            if (shouldRun)
            {
                //This tick runs every 400ms and processes next message to be sent.
                lock (this.UserQueue) //For threadsafety
                {
                    User user;
                    if (this.UserQueue.TryDequeue(out user))            //Pop the next user to be processed.
                    {
                        string message = user.RequestQueue.Dequeue();   //Pop his request
                        this.host.Say(user.Username, message);
                        if (user.RequestQueue.Count > 0)                //If user has more messages waiting to be processed
                        {
                            this.UserQueue.Enqueue(user);               //Add him at the end of the userqueue
                        }
                        else
                        {
                            this.Users.Remove(user.Username);           //User has no more messages, we can safely remove him from the user list
                        }
                    }
                }
            }
        }
    }
}
